#include <math.h>

#include "lca.h"
#include "shaunlib.h"


void LCA(Param &p, LCAParam &lp, Data &d, Output &o)
{

  // ----------------------------------------------------------------------------------
  // IMPLEMENTATION OF LATENT CLASS ANALYSIS FOR GENOMIC CONTROL
  // ----------------------------------------------------------------------------------

  // ----------------------------------------------------------------------------------
  // VARIABLE DEFINITIONS 
  //
  // i  Genotypic configuration
  // j  Latent class
  // l  Locus
  // k  Allele
  //
  // h[j][i]      P(C|R)
  // n[j]         P(C)
  // r[l][j][k]   P(R|C)
  //
  // n_[j]        Count for P(C)
  // r_[l][j][k]  Count for P(R|C)
  //
  // ----------------------------------------------------------------------------------



  o.ml = 0 ; 

  // OUTPUT
  if (!p.min && p.verbose2 && !p.verbose3) 
    cout << "\nk:"<<p.tag<<": Information about repeat E-M convergence";


  for (int ItRepeat=1; ItRepeat <= p.nItRepeat; ItRepeat++)
    {
      
      // ----------------------------------------------------------------------------------
      // DEFINE INITIAL SET P(C|R)
      // ----------------------------------------------------------------------------------

      for (int i=1; i <= d.nconfig; i++)
	{
	  double ftemp = 0;
	  for (int j=1; j < lp.nClass; j++)
	    {
	      double rn = CRandom::rand();
	      rn /= (float)lp.nClass-1;
	      lp.h[j][i] = rn;
	      ftemp += rn;
	    }
	  lp.h[lp.nClass][i] = 1 - ftemp;
	  
	  // ----------------------------------------------------------------------------------
	  // IMPLEMENT FIX-TO-CLASS OPTION
	  // ----------------------------------------------------------------------------------
	  
	  // if option on, if configuration is fixed
	  if ( (p.fix_to_class) && (d.fixconfig2c(i) > 0)  ) 
	    for (int j=1; j <= lp.nClass; j++)
	      {
		if (j==d.fixconfig2c(i)) lp.h[j][i] = 1; 
		if (j!=d.fixconfig2c(i)) lp.h[j][i] = 0;
	      }
	}


      // ----------------------------------------------------------------------------------
      // COUNT TOTAL # OF ALLELES IN SAMPLE
      // ----------------------------------------------------------------------------------

      int totAlleles = 0;
      for (int i=1 ; i <= d.nInd; i++)
	for (int l=1 ; l <= d.nLoci ; l++)
	  if (d.config2(i)(l)==2) totAlleles += 1; 
	  else totAlleles += 2; 


      // ----------------------------------------------------------------------------------
      // PUT SAMPLE ALLELE FREQUENCIES INTO P(R|C)
      // ----------------------------------------------------------------------------------

      // -----------------------------------------
      // Clear derived class individual counts
      // -----------------------------------------

      for (int j=1; j <= lp.nClass; j++)
	lp.n_[j] = 0;

      // -----------------------------------------
      // Clear non-specific allele ancestral allele counts
      // -----------------------------------------

      for (int l=1; l <= d.nLoci; l++) 
	for (int jj=1; jj <= lp.nAncClass; jj++)
	  lp.anc_n_[l][jj] = 0; 

      // -----------------------------------------
      // Generate starting values for ancestral classes
      // -----------------------------------------

      for (int l=1; l <= d.nLoci; l++)
	{
	  double l_sum = 0;

	  // Calculate total # of alleles for that locus
	  int denom = 0;
	  for (int i=1 ; i <= d.nInd; i++)
	    if (d.config2(i)(l)==2) denom += 1; 
	    else denom += 2; 
	  
	  CVector start_freqs(d.nallele(l));

	  // For all non-missing alleles
	  for (int k=3; k <= d.nallele(l); k++)
//	      l_sum += (double)d.fallele(l)(k)/denom; // ORIGINAL
	      {
		  start_freqs[k] += (double)d.fallele(l)(k)/denom + (CRandom::rand()/10); // TEMPORARY
		  if (start_freqs[k] <= 0) start_freqs[k] == 0.001;
		  if (start_freqs[k] >= 1) start_freqs[k] == 0.999;
		  l_sum += start_freqs[k];
	      }

	  for (int jj=1; jj <= lp.nAncClass; jj++)    
	    // For all non-missing alleles
	    for (int k=3; k<=d.nallele(l); k++)
	      {
//		lp.anc_r[l][jj][k] = ((double)d.fallele(l)(k)/denom)/l_sum; ORIGINAL
		  lp.anc_r[l][jj][k] = start_freqs[k]; // TEMPORARY
		lp.anc_r_[l][jj][k] = 0;
	      }
	}
      
      // -----------------------------------------
      // Calculate starting values for derived classes
      // -----------------------------------------

      // clear r frequencies
      for (int l=1; l <= d.nLoci; l++)
	for (int j=1; j <= lp.nClass; j++)
	  {
	    // Set frequency of missing alleles to 1 
	    // so that they are ignored in Step 2 
	    // likelihood calculations 
	    lp.r[l][j][1] = 1; 
	    lp.r[l][j][2] = 1;
	    for (int k=3; k <= d.nallele(l); k++)
	      lp.r[l][j][k] = 0;
	  }

      // calculated weighted sum for admixed allele frequencies
      for (int jj=1; jj <= lp.nAncClass; jj++)	
	for (int l=1; l <= d.nLoci; l++)
	  for (int k=3; k<=d.nallele(l); k++)
	    for (int j=1; j <= lp.nClass; j++)
	      lp.r[l][j][k] += lp.anc_r[l][jj][k] * lp.cls[j][jj] ;
      


      // ----------------------------------------------------------------------------------
      // BEGIN ITERATING THE E-M ALGORITHM  
      // ----------------------------------------------------------------------------------

      // Set iteration counter to zero
      int ii = 0;
      
      // Set tolerance for between-iteration ML change 
      long double ml = 0;
      long double ml0 = 0;

      do
	{
	  
	  // increase iteration counter;
	  ii++;   

	  // ----------------------------------------------------------------------------------
	  // STEP ONE : COUNTING ALLELES (R_) AND INDIVIDUALS (N_) 
	  // ----------------------------------------------------------------------------------


	  // -----------------------------------------
	  // Begin looping over all configurations
	  // -----------------------------------------

	  for (int i=1; i <= d.nconfig; i++) 
	    {

	      // -----------------------------------------
	      // Count individuals to derived classes
	      // -----------------------------------------

	      for (int j=1; j <= lp.nClass; j++) 
		lp.n_[j] += lp.h[j][i] * (float)d.fconfig(i);


	      // -----------------------------------------
	      // Loop over each locus 
	      // -----------------------------------------

	      for (int l=1; l <= d.nLoci; l++) 
		{
		  
		  // -----------------------------------------
		  // For a specific allele k at locus l 
		  // for a specific derived class j
		  // for a specific ancestral class jj 
		  // calculate the count quantity which 
		  // will be added to ancestral class jj 
		  // -----------------------------------------
		  
		  
		  // -----------------------------------------
		  // For each ancestral class
		  // -----------------------------------------
		  for (int jj=1; jj <= lp.nAncClass; jj++) 
		    {
		  
		      // -----------------------------------------
		      // Reset contribution for all alleles
		      // -----------------------------------------
		      for (int k=3; k<=d.nallele(l); k++)
			lp.contrib[k] = 0; 
		      
		      // -----------------------------------------
		      // Sum over derived classes, pure and admix1
		      // -----------------------------------------
		      for (int j=1; j <= lp.nPureClass+(int)lp.nAdmixClass2; j++) 
			{

			  // -----------------------------------------
			  // For each allele individually
			  // -----------------------------------------
			  for (int k=3; k <= d.nallele(l); k++)
			    {
			      double denom = 0;

			      for (int jj2=1; jj2 <= lp.nAncClass; jj2++)
				denom += lp.cls[j][jj2] * lp.anc_r[l][jj2][k];
			      if (denom==0) denom=1;
			      
			      long double temp_contrib = lp.h[j][i] * ((lp.cls[j][jj] * lp.anc_r[l][jj][k])/denom);
			      lp.contrib[k] += temp_contrib;
			    } // next k
			} // next pure or admix1 derived class

	      
		      // -----------------------------------------
		      // Begin counting for this ancestral class jj
		      // -----------------------------------------

		      // -----------------------------------------
		      // Consider if allele is nonmissing...
		      // -----------------------------------------
		  
		      if (d.config1(i)(l)!=1) 
			{
			
			  // -----------------------------------------
			  // ...for diploid organisms
			  // -----------------------------------------
			  if (!p.haploid && d.config2(i)(l)!=2)
			    {

			      // allele-specific ancestral class count
			      lp.anc_r_[l][jj][d.config1(i)(l)] 
				+= lp.contrib[d.config1(i)(l)] * (float)d.fconfig(i) ;
			      
			      lp.anc_r_[l][jj][d.config2(i)(l)] 
				+= lp.contrib[d.config2(i)(l)] * (float)d.fconfig(i) ;

						      
			      // allele-nonspecific ancestral class count
			      lp.anc_n_[l][jj]
				+= lp.contrib[d.config1(i)(l)] * (float)d.fconfig(i) ;

			      lp.anc_n_[l][jj]
				+= lp.contrib[d.config2(i)(l)] * (float)d.fconfig(i) ;
      			      
			    }
			  
			  // -----------------------------------------
			  // ...for haploid organisms
			  // -----------------------------------------
			  if (p.haploid || d.config2(i)(l)==2)
			    {

			      // allele-specific ancestral class count
			      lp.anc_r_[l][jj][d.config1(i)(l)] 
				+= lp.contrib[d.config1(i)(l)] * (float)d.fconfig(i) ;

			      // allele-nonspecific ancestral class count
			      lp.anc_n_[l][jj] 
				+= lp.contrib[d.config1(i)(l)] * (float)d.fconfig(i) ;
			    }

			}
		      // -----------------------------------------
		      // Otherwise consider missing alleles
		      // -----------------------------------------
		      else  
			{
			  
			  // (REMEMBER : MIGHT BREAK DOWN FOR ADMIX2)

			  // -----------------------------------------
			  // Consider each possible nonmissing allele...
			  // -----------------------------------------
			  for (int k=3; k <= d.nallele(l); k++)
			    {
					     			      
			      // -----------------------------------------
			      // ...for each derived class
			      // -----------------------------------------
  			      for (int j=1; j <= lp.nClass; j++)
				{

				  long double mr = lp.h[j][i] * lp.r[l][j][k];

				  // -----------------------------------------
				  // ...and count for diploid organisms
				  // -----------------------------------------

				  if (!p.haploid && d.config2(i)(l)!=2)
				    {
				      lp.anc_r_[l][jj][k] += 2 * lp.contrib[k] * (float)d.fconfig(i) * mr * lp.cls[j][jj] ;
				      lp.anc_n_[l][jj] += 2 * lp.contrib[k] * (float)d.fconfig(i) * mr * lp.cls[j][jj];
				    }
			      
				  // -----------------------------------------
				  // ...and count for haploid organisms
				  // -----------------------------------------
				  if (p.haploid || d.config2(i)(l)==2)
				    {				    
				      lp.anc_r_[l][jj][k] += lp.contrib[k] * (float)d.fconfig(i) * mr * lp.cls[j][jj];
				      lp.anc_n_[l][jj] += lp.contrib[k] * (float)d.fconfig(i) * mr * lp.cls[j][jj];
				    }
				  
				  // -----------------------------------------
				  // Finish looping over all missing alleles
				  // -----------------------------------------
				}
		      
			      // -----------------------------------------
			      // Finish looping over all missing alleles
			      // -----------------------------------------
			    }
			  
			  // -----------------------------------------
			  // End of IF MISSING exception
			  // -----------------------------------------
			} 
		      
		    } // next ancestral class
		  
		} // next locus
	      
	    } // next configuration
	  

	  // ----------------------------------------------------------------------------------
	  // CALCULATE ALLELE FREQS FOR ADMIXED CLASSES
	  // ----------------------------------------------------------------------------------

	  // clear r frequencies
	  for (int l=1; l <= d.nLoci; l++)
	    for (int j=1; j <= lp.nClass; j++)
	      for (int k=3; k <= d.nallele(l); k++)
		lp.r[l][j][k] = 0;

	  // calculated weighted sum for admixed allele frequencies
	  for (int jj=1; jj <= lp.nAncClass; jj++)	
	    for (int l=1; l <= d.nLoci; l++)
	      {
		for (int k=3; k <= d.nallele(l); k++)
		  {
		    // calculate ancestral class allele frequencies
		    if (lp.anc_n_[l][jj] > 0)
		      lp.anc_r[l][jj][k] = lp.anc_r_[l][jj][k] / lp.anc_n_[l][jj];
		    else
		      lp.anc_r[l][jj][k] = 0;
		    // calculate derived class allele frequencies
		    for (int j=1; j <= lp.nClass; j++)
		      lp.r[l][j][k] += lp.anc_r[l][jj][k] * lp.cls[j][jj] ;
		    
		    // reset ancestral class allele counts
		    lp.anc_r_[l][jj][k] = 0 ;
		  }
		// reset individual count of ancestral classes
		// (locus specific -- maybe ?? check )
		lp.anc_n_[l][jj] = 0; 
	      }
	  
	  // ----------------------------------------------------------------------------------
	  // CALCULATE PRIOR FREQUENCIES P(C) FROM COUNTS
	  // ----------------------------------------------------------------------------------
	  
	  for (int j=1; j <= lp.nClass; j++)
	    {
	      lp.n[j] = lp.n_[j]/(float)d.nInd;
	      lp.n_[j] = 0; // reset count 
	    }



	  // ----------------------------------------------------------------------------------
	  // STEP TWO : CALCULATE P(R) - LIKELIHOOD OF THAT CONFIGURATION
	  // ----------------------------------------------------------------------------------
	
	  long double ftemp;
	  ml0 = ml;
	  ml = 0;
	  for (int i=1; i<=d.nconfig; i++)       // for each configuration
	    {
	      lp.f[i] = 0;                        // set to 0
	      for (int j=1; j <= lp.nClass; j++)  // for each class
		{
		  ftemp = lp.n[j];
		  for (int l=1; l <= d.nLoci; l++) // for each response/loci
		    {
		      if (!p.haploid && d.config2(i)(l)!=2)
			{
			  // Is the class Type II new admixture
			  if (j > lp.nPureClass+(int)lp.nAdmixClass2)
			    {
			      // CHECK THIS ... 
// 			      int temp_class = j - (lp.nPureClass+(int)lp.nAdmixClass2) - 1;
// 			      double prob1 = lp.r[l][lp.Admix2Class1[temp_class]][d.config1(i)(l)]
// 				* lp.r[l][lp.Admix2Class2[temp_class]][d.config2(i)(l)];
// 			      double prob2 = lp.r[l][lp.Admix2Class1[temp_class]][d.config2(i)(l)]
// 				* lp.r[l][lp.Admix2Class2[temp_class]][d.config2(i)(l)];
// 			      ftemp *= 0.5*(prob1+prob2);
			    } 
			  else // for normal classes 
			    {
				ftemp *= lp.r[l][j][d.config1(i)(l)];
				ftemp *= lp.r[l][j][d.config2(i)(l)]; 
				// for heterozygotes
				if (d.config1(i)(l)!=d.config2(i)(l)) ftemp *= 2;

			    }
			}
		      if (p.haploid || d.config2(i)(l)==2)
			{
			  ftemp *= lp.r[l][j][d.config1(i)(l)];
			}
		    }
		  lp.f[i] += ftemp;
		}
	      ml += log(lp.f[i]) * (float)d.fconfig(i);
//	      ml *= lp.f[i]; // Assume GROUP GENO option is always off, for now
	    }

	  // ----------------------------------------------------------------------------------
	  // STEP TWO : CALCULATE P(R|C)
	  // ----------------------------------------------------------------------------------

	  // ------------------------------------------
	  // For each configuration
	  // ------------------------------------------
	  for (int i=1; i <= d.nconfig; i++) 
	    {

	      // ------------------------------------------
	      // For each derived class
	      // ------------------------------------------
	      for (int j=1; j <= lp.nClass; j++)  
		{

		  // ------------------------------------------
		  // Set posterior prob P(C|R) to P(C)
 		  // ------------------------------------------
		  ftemp = lp.n[j];

		  // ------------------------------------------
		  // Consider each locus
		  // ------------------------------------------
		  for (int l=1; l <= d.nLoci; l++)    
		    {

		      // ------------------------------------------
		      // For diploid organisms
		      // ------------------------------------------
		      if (!p.haploid && d.config2(i)(l)!=2)
			{

			  // ------------------------------------------
			  // For Admix2 type derived classes
			  // ------------------------------------------
			  if (j > lp.nPureClass+(int)lp.nAdmixClass2)
			    {
// 			      int temp_class = j - (lp.nPureClass+(int)lp.nAdmixClass2) -1;
			      
// 			      double prob1 = lp.r[l][lp.Admix2Class1[temp_class]][d.config1(i)(l)]
// 				* lp.r[l][lp.Admix2Class2[temp_class]][d.config2(i)(l)];
			      
// 			      double prob2 = lp.r[l][lp.Admix2Class1[temp_class]][d.config2(i)(l)]
// 				* lp.r[l][lp.Admix2Class2[temp_class]][d.config1(i)(l)];
			      
// 			      lp.h[j][i] *= 0.5*(prob1+prob2);

			      // Do we still need to do this?
			      // (and above in P(G=i) ?? )
			      //if (config[i][l][1]!=config[i][l][2]) h[j][i] *= 2;

			    } 
			  else
			 
			    // ------------------------------------------
			    // For non-Admix2 type derived classes
			    // ------------------------------------------
			    {
			      
				// ------------------------------------------
				// Homozygotes
				// ------------------------------------------
				ftemp  *= lp.r[l][j][d.config1(i)(l)];
				ftemp  *= lp.r[l][j][d.config2(i)(l)];
				    
				// ------------------------------------------
				// Heterozygotes
				// ------------------------------------------
				if (d.config1(i)(l)!=d.config2(i)(l)) ftemp *= 2;

			    }
			}

		      // ------------------------------------------
		      // Otherwise, for haploid organsisms
		      // ------------------------------------------
		      if (p.haploid || d.config2(i)(l)==2)
			ftemp *= lp.r[l][j][d.config1(i)(l)];

		      // ------------------------------------------
		      // Loop back to next locus
		      // ------------------------------------------
		    }

		  // ------------------------------------------
		  // Conditional likelihood of configuration
		  // ------------------------------------------
		  ftemp /= lp.f[i];
		  lp.h[j][i] = ftemp;

		  // ----------------------------------------------------------------------------------
		  // IMPLEMENT FIX-TO-CLASS OPTION
		  // ----------------------------------------------------------------------------------
		  
		  // if option on, if configuration is fixed
		  if ((p.fix_to_class) && (d.fixconfig2c(i) > 0) ) 
		    {
		      if (j==d.fixconfig2c(i)) lp.h[j][i] = 1; 
		      if (j!=d.fixconfig2c(i)) lp.h[j][i] = 0;
		    } 

		  // ------------------------------------------
		  // Consider next class
		  // ------------------------------------------
		} 

	      // ------------------------------------------
	      // Consider next configuration
	      // ------------------------------------------
	    } 

	  // ------------------------------------------
	  // Verbose(2) output
	  // ------------------------------------------
	  if (!p.min && p.verbose3)
	    {
		cout << "\nOPT:"<<p.tag<<": LL = " << ml << " change = " << (ml0-ml) ; // << ", " << ml/ml0;
	      cout << "   ";
	      for (int i0=1; i0 <= lp.nClass; i0++) cout << "\t P(C="<<i0<<")="<<lp.n[i0];
	    }

	  // ------------------------------------------
	  // Have we converged yet?
	  // ------------------------------------------

	 
	}  while (((ml0-ml) < -p.tol) || ((ml0-ml) > p.tol));
      // while ( (ml/ml0) < 1 || (ml/ml0) > 1.01 ) ; 

      // ----------------------------------------------------------------------------------
      // END OF E-M ALGORITHM -- SUCCESSFULLY CONVERGED
      // ----------------------------------------------------------------------------------

      if (!p.min && p.verbose2)
	{
	  if (p.verbose3) cout << "\nk:"<<p.tag<<": Information about repeat E-M convergence";
	  cout << "\nk:"<<p.tag<<":\t" << lp.nClass << "\t" 
	       << ii << "\t" << -2*ml ;
	  for (int j=1; j <= lp.nClass; j++)
	    cout << "\tP(C=" << j << ")=" << lp.n[j];
	  if (p.verbose3) cout << "\n";
	}

      if (!p.min && !p.verbose2 && p.verbose3) cout << "\n";

      // ----------------------------------------------------------------------------------
      // HAS THIS CONVERGENCE GIVEN THE BEST ESTIMATES ? 
      // ----------------------------------------------------------------------------------
     
      if ((ItRepeat == 1) || (ml > o.ml))
	{
	  o.ml = ml ; 
	  o.ii = ii ; 
	  for (int j=1; j <= lp.nClass; j++) 
	    {
	      o.n[j] = lp.n[j]; 
	      for (int i=1; i <= d.nconfig; i++)
		o.h[j][i] = lp.h[j][i] ;  
	      for (int l=1; l <= d.nLoci; l++)
		for (int k=1; k<=d.nallele(l); k++)
		  o.r[l][j][k] = lp.r[l][j][k] ; 
	    }
	}


      // ----------------------------------------------------------------------------------
      // RETURN FOR MORE ATTEMPTS AT CONVERGENCE (ItRepeat LOOP)
      // ----------------------------------------------------------------------------------

      if (!p.min && !p.distance) (ItRepeat % 5 == 0) ? cerr << "|" : cerr << "-";
    }

  if (!p.min && !p.distance) cerr << "*\n";

  // ----------------------------------------------------------------------------------
  // CALCULATE AIC FROM LIKELIHOOD
  // ----------------------------------------------------------------------------------
  
  // prior probabilities for all classes
  lp.npar = lp.nClass-1;  

  // class-specific allele freqs for ancestral classes only
  // minus 3 -i.e. minus 1, 
  // then another as allele '1' is missing allele
  // then another as allele '2' is Xchr allele
  for (int i0=1; i0<=d.nLoci; i0++)
    lp.npar += lp.nAncClass*(d.nallele(i0)-3);


  // ----------------------------------------------------------------------------------
  // FIND 'BEST-FIT' CLASS ESTIMATES (HIGHEST POSTERIOR PROBABILITY)
  // ----------------------------------------------------------------------------------
  
  for (int i=1; i <= d.nInd; i++)
    {
      lp.indClass(i) = 1;
      double indMax = o.h[1][d.ic(i)];
      for (int j=2; j <= lp.nClass; j++)
	if (o.h[j][d.ic(i)] > indMax) 
	  {
	    lp.indClass(i) = j;
	    indMax = o.h[j][d.ic(i)];
	  }
    }
    
}







