#include "shaunlib.h"
#include "param.h"
#include "data.h"

#include <iomanip>
#include <fstream>
#include <time.h>
#include <stdlib.h>

using namespace std;

// ----------------------------------------------------------------------------------
// GET PARAM FILE FROM STANDARD INPUT
// ----------------------------------------------------------------------------------

Param::Param()
{
  // Default parameter values
  title = "";
  tag = "";
  haploid = false;
  hwd = false; 
  entropy = false;
  locinfo = false;
  cslocinfo = false;
  structure_format = false;
  fix_to_class = false;
  admix_res = 1;         
  random_seed = 12345678;
  datafile = "data.raw"; 
  header_row = false;
  fix_n_loci = false;
  group_geno = false; 
  tol = 0.0001;       
  verbose1 = false;
  verbose2 = false;
  verbose3 = false;
  min = false;
  debug = false;
  distance = false;
  nPheno = 0;  
  nAllele = 5;
  valX = -1 ; 
  xdata = false;
  valMissing = -9;    
  nItRepeat = 10;     
  admix_res = 1; 
}

LCAParam::LCAParam()
{
  nPureClass = 1; 
  nAdmixClass = 0;
  nAdmix1Class = 0;
  nAdmix2Class = 0; 
  setPure = false;
  setAuto = false;
  setA1 = false;
}

LCAParam::~LCAParam()
{
/// 
}

void Param::get_param(LCAParam &lp, Data &d)
{

  // get title line
  getline(cin,title,'\n');

  string comm;

  do {
    // read command
    cin >> comm;
   // ignore if # comment #
    if (comm=="#") { string dummy; while (dummy!="#") cin >> dummy; }
    // parse commands
    if (comm=="RAND")         cin >> random_seed;
    if (comm=="DATAFILE")     cin >> datafile;
    if (comm=="STRUCTURE")    structure_format = true;  
    if (comm=="HEADER")       header_row = true;
    if (comm=="HAPLOID")      haploid = true; 
    if (comm=="HWD")          hwd = true;
    if (comm=="ENTROPY")      entropy = true;
    if (comm=="LOCINFO")      locinfo = true;
    if (comm=="CSLOCINFO")    cslocinfo = true;
    if (comm=="FIX2CLASS")    fix_to_class = true;  
    if (comm=="GROUP")        group_geno = true;  
    if (comm=="PHENO")        cin >> nPheno;
    if (comm=="REPEAT")       cin >> nItRepeat;
    if (comm=="TAG")          cin >> tag;
    if (comm=="TOL")          cin >> tol;
    if (comm=="MIN")          min = true; 
    if (comm=="VERBOSE2")     verbose2 = true;
    if (comm=="VERBOSE1")     verbose1 = true;
    if (comm=="VERBOSE3")     verbose3 = true;
    if (comm=="DEBUG")        debug = true;
    if (comm=="DISTANCE")     distance = true;
    if (comm=="XCHR")         cin >> valX; 
    if (comm=="MISSING")      cin >> valMissing;

    if (comm=="CLASS")       { cin >> lp.nPureClass; lp.setPure = true; }

    if (comm=="A1AUTO")       
      {
	if (lp.setA1) 
	  error("Sorry -- cannot use A1CLASS and A1AUTO together...");

	cin >> lp.nAdmixClass; 
	lp.setAuto = true; 
      }

    if (comm=="A1LEVEL")      cin >> admix_res;

    if (comm=="A1CLASS")
      {
	lp.setA1 = true;

	// Must already know lp.nPureClass;
	if (!lp.setPure) 
	  error("Must specify pure classes before using A1CLASS.");

	if (lp.setAuto)
	  error("Sorry -- cannot use A1CLASS and A1AUTO together...");
	
	cin >> lp.Admix1Class1[lp.nAdmix1Class];
	cin >> lp.Admix1Class2[lp.nAdmix1Class];
	lp.nAdmix1Class++;
      }


    if (comm=="A2CLASS")
      {
	cin >> lp.Admix2Class1[lp.nAdmix2Class];
	cin >> lp.Admix2Class2[lp.nAdmix2Class];
	lp.nAdmix2Class++;
      }

    if (comm=="ALLELE") cin >> nAllele;

    // reset value of comm
    comm="";

  } while (!cin.eof());

  if (group_geno && fix_to_class)
    error("Cannot GROUP genotypes and FIX2CLASS simultaneously.");

  if (random_seed == 0) CRandom::srand(time(0));
  else CRandom::srand(random_seed);

  cout << setprecision(4) << setiosflags(ios::fixed | ios::showpoint);

}



void Param::get_loci_ind(LCAParam &lp, Data &d)
{

  // ----------------------------------------------------------------------------------
  // CALCULATE NUMBER OF LOCI, ALLELES AND INDIVIDUALS IN DATA
  // ----------------------------------------------------------------------------------

  CData d2;
  d2.field_delimit(" ,\t");
  d2.line_delimit("\n");
  d2.header = false;
  d2.import(datafile);
  
  // # loci
  nLoci = d2.getVar();
  nLoci--;  // One column always ID code
  nLoci -= nPheno; // Subtract any phenotype columns
  if (fix_to_class) nLoci--; // If f2c column specified
  if ((!haploid) && (!structure_format)) nLoci /= 2; 

  nInd= d2.getInd(); 
  // If haploid organisms in Structure file mode  
  if ((!haploid) && (structure_format)) nInd /= 2; 
  
  // Data nInd reflects actual # of individuals
  // Param nInd reflects individuals + missing individual
  d.nInd = nInd;
  d.nLoci = nLoci;

  // Copy for deallocation of ***r
  lp.nAllele = nAllele;
  lp.nLoci = nLoci;

  // Copy
  lp.admix_res = admix_res;

}



// ----------------------------------------------------------------------------------
// CALCULATE NUMBER OF LATENT CLASSES REQUIRED
// ----------------------------------------------------------------------------------
void Param::get_class(LCAParam &lp)
{
  
  lp.nClass = lp.nPureClass;
  lp.nAncClass = lp.nPureClass + lp.nAdmixClass;

  // Automatically derived A1 classes
  if (lp.setAuto)
    {
      lp.nAdmixClass2 = 1;
      for (int i = 0 ; i <= admix_res-1 ; i++) 
	lp.nAdmixClass2 *= (lp.nAdmixClass + i) / float(1 + i);
      
      lp.nClass += (int)lp.nAdmixClass2;
    }
  else
    lp.nAdmixClass2 = lp.nAdmix1Class;  // # derived I classes.

  // Manually specified Type I and Type II admixture
  lp.nClass += lp.nAdmix1Class + lp.nAdmix2Class ;

}



// ----------------------------------------------------------------------------------
// ALLOCATE STORAGE
// ----------------------------------------------------------------------------------
void Param::get_storage(LCAParam &lp, Data &d)
{
  
  d.d1.setRow(d.nInd); d.d1.setCol(nLoci);
  d.d2.setRow(d.nInd); d.d2.setCol(nLoci);
  d.id().resize(d.nInd);
  d.ic.setCol(d.nInd);
  d.ph.setRow(d.nInd); d.ph.setCol(nPheno);
  d.fix2c.setCol(d.nInd);
  d.s1.setRow(d.nInd); d.s1.setCol(nLoci);
  d.s2.setRow(d.nInd); d.s2.setCol(nLoci);
  d.c.setRow(nLoci); d.c.setCol(nAllele);
  d.fallele.setRow(nLoci); d.fallele.setCol(nAllele);
  d.nallele.setCol(nLoci);
  

  d.fconfig.setCol(nInd);
  d.config1.setRow(nInd); d.config1.setCol(nLoci); 
  d.config2.setRow(nInd); d.config2.setCol(nLoci); 

  lp.cls.setRow(lp.nClass); lp.cls.setCol(lp.nAncClass);
  lp.h.setRow(lp.nClass); lp.h.setCol(nInd);
  lp.indClass.setCol(d.nInd);


  lp.r.setDep(nLoci);
  lp.r.setRow(lp.nClass);
  lp.r.setCol(nAllele);

  lp.r_.setDep(nLoci);
  lp.r_.setRow(lp.nClass);
  lp.r_.setCol(nAllele);

  lp.anc_r.setDep(nLoci);
  lp.anc_r.setRow(lp.nAncClass);
  lp.anc_r.setCol(nAllele);

  lp.anc_r_.setDep(nLoci);
  lp.anc_r_.setRow(lp.nAncClass);
  lp.anc_r_.setCol(nAllele);

  lp.contrib.setCol(nAllele);
  lp.n.setCol(lp.nClass);
  lp.n_.setCol(lp.nClass);
  lp.anc_n_.setRow(nLoci); 
  lp.anc_n_.setCol(lp.nAncClass); 
  lp.f.setCol(nInd);
}

void LCAParam::free_storage()
{
//   mymath::free_mp_3tensor(r,1,nLoci,1,nClass,1,nAllele);
//   mymath::free_mp_3tensor(r_,1,nLoci,1,nClass,1,nAllele);
//   mymath::free_mp_3tensor(anc_r,1,nLoci,1,nAncClass,1,nAllele);
//   mymath::free_mp_3tensor(anc_r_,1,nLoci,1,nAncClass,1,nAllele);
}




// ----------------------------------------------------------------------------------
// CREATE CLASS STRUCTURE FOR ADMIXED CLASS FROM ANCESTRAL CLASSES
// ----------------------------------------------------------------------------------
void LCAParam::make_admix()
{
  // Temporarily forget about the new admixed type II classes
  nClass -= nAdmix2Class;

  ///////////////////////////////////////////////////////////
  // For Pure classes
  for (int i = 1; i <= nClass; i++) 
    for (int j = 1; j <= nPureClass; j++)
      if (i == j) 
	cls[i][j] = 1;
      else
	cls[i][j] = 0; 

  ///////////////////////////////////////////////////////////  
  // For any Admixed I classes -- automatic
  if (nAdmixClass>0)
    {
      // Beginning of admixed classes
      int i = nPureClass+1;

      IntArray z;
      z.setCol(nAdmixClass);

      for (int j=1; j <= nAdmixClass; j++) z(j) = 0;

      for (;;)
	{
	  // get sum of z[]
	  int z_sum = 0;
	  for (int j=1; j <= nAdmixClass; j++) {
	    z_sum += z(j);
	  }

	  // does sum equal the admixture resolution? 
	  if (z_sum == admix_res) // if so, copy elements to main matrix 
	    {
	      for (int j=1; j <= nAdmixClass; j++) 
		cls[i][j+nPureClass] = z(j) / (float)admix_res;
	      // increment counter
 	      i++;
	    }
	  
	  // increment 
	  int temp=1;
	  do 
	    {
	      z(temp)++;
	      if (z(temp) > admix_res) 
		{
		  z(temp) = 0; 
		  temp++;
		}
	      else { break; }
	    }
	  while (temp <= nAdmixClass);
      
	  // done?
	  if (z(nAdmixClass) == admix_res) 
	    {
	      // save last good one and exit 
	      for (int j=1; j <= nAdmixClass; j++) 
		cls[i][j+nPureClass] = z(j) / (float)admix_res;
	      break;
	    }
	}
    }

  // Restore correct number of classes
  nClass += nAdmix2Class;


  ///////////////////////////////////////////////////////////  
  // For any Admixed I classes -- manual


  for (int i = 1 ; i <= nAdmix1Class ; i++ ) 
    {
      cls[i+nPureClass][Admix1Class1[i-1]] = 0.5;
      cls[i+nPureClass][Admix1Class2[i-1]] = 0.5;
    }


  // ----------------------------------------------------------------------------------
  // ADD NEW NUMBER OF TYPE II ADMIXED CLASS
  // ----------------------------------------------------------------------------------

  for (int i = 1 ; i <= nAdmix2Class ; i++ ) 
    {
      cls[i+nPureClass+(int)nAdmixClass2][Admix2Class1[i-1]] = 0.5;
      cls[i+nPureClass+(int)nAdmixClass2][Admix2Class2[i-1]] = 0.5;
    }

}

